package de.hpi.imdbsuche.database;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;

import org.hsqldb.persist.HsqlProperties;
import org.hsqldb.server.Server;

import com.opencsv.CSVReader;

import de.hpi.imdbsuche.entities.DBEntity;
import de.hpi.imdbsuche.entities.imdb.Actor;
import de.hpi.imdbsuche.entities.imdb.Actress;
import de.hpi.imdbsuche.entities.imdb.Genre;
import de.hpi.imdbsuche.entities.imdb.Movie;
import de.hpi.imdbsuche.entities.imdb.Producer;

/**
 * The DatabaseServer starts a HSQLDB server, creates a database and some tables, and inserts data into these tables.
 * (http://hsqldb.org/doc/guide/running-chapt.html)
 */
public class DatabaseServer implements AutoCloseable {
	
	private final String databaseUrl;
	private final String username;
	private final String password;
	
	public String getDatabaseUrl() {
		return this.databaseUrl;
	}

	public String getUsername() {
		return this.username;
	}

	public String getPassword() {
		return this.password;
	}

	public DatabaseServer(String databaseName, String username, String password) throws Exception {

		// Create property values
		String databaseType = "mem:"; //"file:"
		String databaseFolder = new File(".").getAbsolutePath() + File.separator + "database";
		
		String database = databaseType + databaseFolder + File.separator + databaseName;
		this.databaseUrl = "jdbc:hsqldb:" + database;
		this.username = username;
		this.password = password;
		
		// Create the server properties (http://hsqldb.org/doc/guide/ch04.html)
		HsqlProperties properties = new HsqlProperties();
		properties.setProperty("server.database.0", database + ";user=" + this.username + ";password=" + this.password);
		properties.setProperty("server.dbname.0", databaseName);
		properties.setProperty("server.port", "9001");
		properties.setProperty("server.silent", "true");
		properties.setProperty("server.trace", "false");
		properties.setProperty("server.no_system_exit", "true");

		// Create and start the database server
		Server server = new Server();
		server.setProperties(properties);
		server.setLogWriter(null);
		server.setErrWriter(null);
		server.start();
	}
	
	public void loadData() throws SQLException, IOException {

		String dataFolder = new File(".").getAbsolutePath() + File.separator + "data";
		
		DBEntity[] entities = {new Movie(), new Actor(), new Actress(), new Genre(), new Producer()};
		
		// Connect to database
		Connection con = DriverManager.getConnection(this.databaseUrl, this.username, this.password);
		con.setAutoCommit(true);
		
		// Create tables
		for (DBEntity entity : entities)
			this.createTable(con, entity);
		
		// Insert data
		for (DBEntity entity : entities)
			this.loadTable(con, dataFolder, entity);
	}
	
	private void createTable(Connection con, DBEntity entity) throws SQLException {
		
		try (Statement stmt = con.createStatement()) {
			stmt.executeUpdate(entity.getCreateTableSql());
		}
	}
	
	private void loadTable(Connection con, String dataFolder, DBEntity entity) throws FileNotFoundException, IOException, SQLException {
		
		String dataFile = dataFolder + File.separator + entity.getTableName() + ".csv";
		try (PreparedStatement stmt = con.prepareStatement(entity.getInsertIntoSql())) {
			try (CSVReader reader = new CSVReader(new InputStreamReader(new FileInputStream(dataFile), StandardCharsets.UTF_8), ',', '\"')) {
				reader.forEach(record -> this.addToBatch(record, stmt));
			}
			stmt.executeBatch();
		}
	}
	
	private void addToBatch(String[] record, PreparedStatement stmt) {
		try {
			for (int i = 0; i < record.length; i++)
				stmt.setString(i + 1, record[i]);
			stmt.addBatch();
		} catch (SQLException e) {
			throw new RuntimeException(e.getMessage());
		}
	}
	
	@Override
	public void close() throws SQLException {

		// Connect to database
		Connection con = DriverManager.getConnection(this.databaseUrl, this.username, this.password);

		// Write data to disk and shutdown the database server
		Statement statement = con.createStatement();
		statement.execute("SHUTDOWN");
		statement.close();

		// Close the connection to the database
		con.close();
	}
}
